#!/usr/bin/python3

# Copyright 2021 Offensive Security
# SPDX-license-identifier: GPL-3.0-only

import logging
import os
import shutil
import sys
import termios
from contextlib import contextmanager

from snack import (
    Grid,
    SnackScreen,
    Textbox,
)

from .toolkit import (
    KaliCheckboxBar,
    KaliConfirmWindow,
    KaliInfoWindow,
    KaliListboxWindow,
    KaliRadioBar,
    KaliSimpleCheckboxTree,
    KaliTextboxFullSize,
    KaliWindow,
)
from .utils import (
    OpenSSLSetting,
    VirtualizationSetting,
    apply_current_apt_sources_config,
    apply_shell_config,
    apply_shell_config_files,
    apt_update,
    bold,
    change_login_shell,
    get_configured_login_shell,
    get_current_apt_sources_config,
    get_current_shell_config,
    get_effective_login_shell,
    get_shell_config_files,
    install_pkgs,
    is_pkg_installed,
    remove_pkgs,
    run,
)

LOG_FILE = "kali-tweaks-log.txt"
LOG_LEVEL = logging.DEBUG

if "DEBUG" in os.environ:
    # Can't use stdout while the UI is on, so we send logs to a file instead
    logging.basicConfig(filename=LOG_FILE, filemode="w", level=LOG_LEVEL)

logger = logging.getLogger("kali-tweaks")

HEIGHT = 12

# -------- HELPERS -------- #


def get_metapackages(search):
    """Return an array of (pkg, description, installed)"""
    out = []

    # Run the command
    res = run(search)

    # If exit code wasn't expected, quit
    if res.returncode != 0:
        return out

    # For every result
    for line in res.stdout.splitlines():
        try:
            # Example: kali-linux-default - Kali Linux default system
            pkg, desc = line.split(" - ", maxsplit=1)
        except ValueError:
            continue
        # Check to see if installed
        installed = is_pkg_installed(pkg)
        # Record
        out += [(pkg, desc, installed)]

    return out


def get_metapackage_items():
    items = []

    # List the Kali-tools
    tools = get_metapackages(
        "apt-cache search --names-only ^kali-tools | grep -v ' menu$'"
    )
    for pkg, desc, installed in tools:
        short_pkg = pkg.removeprefix("kali-tools-")
        short_desc = desc.removeprefix("Kali Linux ")
        items += [(short_pkg, short_desc, installed)]

    # List the Kali-linux metapackages
    tools = get_metapackages(
        "apt-cache search --names-only ^kali-linux"
        r" | grep -vE '\-(arm|nethunter|core|top10) '"
    )
    for pkg, desc, installed in tools:
        items += [(pkg, desc, installed)]

    return items


# -------- UI HELPERS -------- #

SCREEN = None


class KaliTweaksScreen:
    def __init__(self):
        self.screen = None
        self.terminal_size = (-1, -1)

    def create(self):
        """
        Create a SnackScreen instance.
        """
        if self.screen is not None:
            logger.warning("The screen object already exists, can't create a new one")
            return

        self.screen = SnackScreen()
        # Theming for checkbox and radiobox
        self.screen.setColor("CHECKBOX", "black", "lightgray")
        self.terminal_size = shutil.get_terminal_size()
        logger.debug("Screen created: %d x %d", self.screen.width, self.screen.height)

    def destroy(self):
        """
        Destroy the current SnackScreen instance.
        """
        if self.screen is None:
            logger.warning("The screen object doesn't exist, can't destroy it")
            return

        self.screen.finish()
        self.screen = None
        self.terminal_size = (-1, -1)
        logger.debug("Screen destroyed")

    def prepare(self):
        """
        Prepare the current screen before showing a window. In other words,
        recreate it if the dimensions of the terminal have changed.
        """
        if not self.screen:
            self.create()
            return

        if shutil.get_terminal_size() != self.terminal_size:
            self.destroy()
            self.create()

    def show_info_window(self, level, msg):
        """
        Show an informative message in a window.

        level can be: info, warning, error.
        """
        self.prepare()
        KaliInfoWindow(self.screen, HEIGHT, level, msg)

    def show_confirm_window(self, title, msg):
        """
        Show a confirmation window, for user to say yes or no.

        Returns the button that was chosen by user.
        """
        self.prepare()
        button = KaliConfirmWindow(self.screen, HEIGHT, msg, title=title)
        return button

    def show_window(self, title, content, buttons):
        """
        Show an interactive window, for user to change settings.

        Returns the button that was chosen by user.
        """
        self.prepare()
        window = KaliWindow(self.screen, title, content, buttons)
        button = window.run_once()
        return button

    def show_listbox_window(self, title, items, buttons):
        """
        Show a list in a window. I guess this should be merged with
        show_window() at some point.
        """
        self.prepare()
        button, line = KaliListboxWindow(self.screen, HEIGHT, title, items, buttons)
        return button, line


@contextmanager
def suspend_screen(ending_message=None, wait_enter=False):
    """
    Suspend the UI while running a command.

    It basically destroys the SnackScreen instance, so that the terminal
    becomes available again for our program. Therefore it shows stdin, stdout
    and stderr, just the usual.

    This is useful to show progress for long operations (package installation),
    or when user interaction is required (for example, entering a password).
    """
    SCREEN.destroy()
    yield
    if ending_message:
        print()
        print("┏━(\033[1;31mMessage from Kali developers\033[00m)")
        for line in ending_message.splitlines():
            print("┃ " + line)
        print("┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━")
        print()
        wait_enter = True
    if wait_enter:
        termios.tcflush(sys.stdin, termios.TCIOFLUSH)
        input(bold("> Press Enter to continue..."))


def say_reload_terminal(first_line):
    msg = f"{first_line}\n"
    msg += "Please close and restart all your terminal instances"
    msg += " for the change to take effect."
    SCREEN.show_info_window("info", msg)


def say_log_out(first_line):
    msg = f"{first_line}\n"
    msg += "Please log out and log back in"
    msg += " for the change to take effect."
    SCREEN.show_info_window("info", msg)


# -------- CLASSES -------- #


class KaliShellDefaultLoginContent(Grid):
    def __init__(self, config):
        super().__init__(1, 2)
        used_height = 0

        # Get the values we care about
        shell = config.get("shell")

        # Build the form with all the settings
        heading = "Default Login Shell to use:"
        items = [
            ("Bash", "Bourne Again SHell", "bash"),
            ("ZSH", "Z SHell", "zsh"),
        ]
        radioboxes = KaliRadioBar(heading, items, shell)
        self.setField(radioboxes, 0, 0, growx=1)
        used_height += 1 + len(items)

        vertical_padding = Textbox(1, HEIGHT - used_height, "")
        self.setField(vertical_padding, 0, 1, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.radioboxes = radioboxes

    def getSelection(self):
        selected = self.radioboxes.getSelection()
        config = {}
        config["shell"] = selected
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            if new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


class KaliShellPromptContent(Grid):
    def __init__(self, config):
        super().__init__(1, 3)
        used_height = 0

        # Get the values we care about
        newline = config.get("newline", True)
        prompt = config.get("prompt", "twoline")

        # Build the form with all the settings
        heading = "Prompt Style:"
        items = [
            ("Two Lines", "Dual blue line prompt", "twoline"),
            ("One Line", "Single blue line prompt", "oneline"),
            ("BackTrack", "Legacy BackTrack red prompt", "backtrack"),
        ]
        radioboxes = KaliRadioBar(heading, items, prompt)
        self.setField(radioboxes, 0, 0, growx=1)
        used_height += 1 + len(items)

        heading = "Prompt Settings:"
        items = [
            ("Newline", "Add a new line between output and prompt", newline),
        ]
        checkboxes = KaliCheckboxBar(heading, items)
        self.setField(checkboxes, 0, 1, growx=1)
        used_height += 1 + len(items)

        vertical_padding = Textbox(1, HEIGHT - used_height, "")
        self.setField(vertical_padding, 0, 2, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.radioboxes = radioboxes
        self.checkboxes = checkboxes

    def getSelection(self):
        config = {}
        config["prompt"] = self.radioboxes.getSelection()
        config["newline"] = self.checkboxes.getSelection()[0]
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            # WARNING! No guarantee that orig config had all keys.
            # At this moment, keys might be missing.
            if key not in orig_config:
                changes[key] = new_config[key]
            elif new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


class KaliNetworkRepoContent(Grid):
    def __init__(self, config):
        super().__init__(1, 2)
        used_height = 0

        # Get the values we care about
        bleeding_edge = config["bleeding-edge"]
        experimental = config["experimental"]

        # Build the form with all the settings
        heading = "Additional Kali repositories:"
        items = [
            (
                "bleeding-edge",
                "Automatically packaged and potentially unstable",
                bleeding_edge,
            ),
            (
                "experimental",
                "Staging area for work-in-progress packages",
                experimental,
            ),
        ]
        checkboxes = KaliCheckboxBar(heading, items)
        self.setField(checkboxes, 0, 0, growx=1)
        used_height += 1 + len(items)

        vertical_padding = Textbox(1, HEIGHT - used_height, "")
        self.setField(vertical_padding, 0, 1, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.checkboxes = checkboxes

    def getSelection(self):
        selected = self.checkboxes.getSelection()
        config = {}
        config["bleeding-edge"] = selected[0]
        config["experimental"] = selected[1]
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            if new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


class KaliGenericContent(Grid):
    # XXX Does not support KaliCheckboXBar yet, only KaliRadioBar
    def __init__(self, settings):
        super().__init__(1, 3)

        self._orig_config = settings
        self._widgets = []

        # No setting to show?
        if not settings:
            msg = r"No setting in this section ¯\_(ツ)_/¯"
            widget = KaliTextboxFullSize(HEIGHT, msg)
            self.setField(widget, 0, 0, growx=1)
            return

        # Iterate on settings and make widgets
        used_height = 0
        for item in settings:

            heading = item["desc"] + ":"
            choices = item["choices"]
            value = item["value"]
            widget = None
            if item["type"] == "radio":
                widget = KaliRadioBar(heading, choices, value)
            # How do handle values?
            # elif item["type"] == "check":
            #    widget = KaliCheckboxBar(heading, items)
            else:
                raise NotImplementedError

            self._widgets.append(widget)
            used_height += 1 + len(choices)

        # Add the widgets
        for index, widget in enumerate(self._widgets):
            self.setField(widget, 0, index, growx=1)

        # Add vertical padding
        padding = HEIGHT - used_height
        if padding <= 0:
            return
        widget = Textbox(1, padding, "")
        self.setField(widget, 0, len(self._widgets), growx=1)

    def getOrigSelection(self):
        """
        Return the original selection
        """
        data = {}
        for setting in self._orig_config:
            setting_name = setting["name"]
            data[setting_name] = setting.value
        return data

    def getCurrentSelection(self):
        """
        Return the current selection, eg:

        {  "my-radiobar-setting": "selected-item",
           "my-checkboxbar-setting": [True, False, True, …],
        }
        """
        data = {}
        for setting, widget in zip(self._orig_config, self._widgets):
            setting_name = setting["name"]
            data[setting_name] = widget.getSelection()
        return data

    def getChanges(self):
        changes = {}
        for setting, widget in zip(self._orig_config, self._widgets):
            setting_name = setting["name"]
            orig = setting["value"]
            new = widget.getSelection()
            if orig != new:
                changes[setting_name] = new

        logger.debug("changes: %s", changes)

        return changes


# -------- SECONDARY SCREENS -------- #


def do_shell_change_default_login():
    title = "Default Login Shell"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get effective and configured login shells, make sure there's no change pending
    configured_shell = get_configured_login_shell()
    effective_shell = get_effective_login_shell()
    if configured_shell != effective_shell:
        msg = f"Your shell is configured to be {configured_shell},"
        msg += f" however at the moment it is effectively {effective_shell}.\n"
        msg += "It seems that you have changes pending,"
        msg += " please log out and log back in for these changes to take effect."
        SCREEN.show_info_window("warning", msg)
        return

    # Prepare a config
    config = {}
    config["shell"] = effective_shell

    # Show the window with all the settings
    content = KaliShellDefaultLoginContent(config)
    btn = SCREEN.show_window(title, content, buttons)
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    old_shell = config.get("shell", "unknown")
    new_shell = changes["shell"]
    msg = ""
    msg += f"Your current login shell is {old_shell}.\n"
    msg += f"Changing it to {new_shell}.\n"
    msg += "Proceed?"

    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return

    # Do the actual job
    try:
        with suspend_screen():
            change_login_shell(new_shell)
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))
        return

    # One last notification
    say_log_out("Your login shell was changed!")


def do_shell_reset():
    configs = get_shell_config_files()

    title = "Reset Shell Config"
    msg = "The following config files will be replaced by Kali's default:\n"
    for default, current in configs:
        msg += f"* {current} (replaced with {default})\n"
    msg += "WARNING: Your changes in those files, if any, will be lost.\n"
    msg += "Proceed?"

    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return

    # Do the actual job
    try:
        with suspend_screen(wait_enter=True):
            apply_shell_config_files(configs)
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))
        return

    # One last notification
    say_reload_terminal("Your shell config files were reset!")


def do_shell_change_prompt():
    title = "Command Prompt"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get the current shell config
    config = get_current_shell_config()

    # Couldn't get the config?
    if config is None:
        msg = "Failed to parse shell configuration!\n"
        msg += "This is expected if your shell configuration files are"
        msg += " too old, and all you need to do is to reset it to the"
        msg += " latest default.\n"
        msg += "Do you want to reset your shell configuration files?"

        btn = SCREEN.show_confirm_window(title, msg)
        if btn == "ok":
            return "go-to-reset"
        else:
            return

    # Show the window with all the settings
    content = KaliShellPromptContent(config)
    btn = SCREEN.show_window(title, content, buttons)
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    msg = ""
    if "prompt" in changes:
        new = changes["prompt"]
        msg += f"Setting prompt to '{new}'.\n"
    if "newline" in changes:
        if changes["newline"] is True:
            msg += "Enabling new line before prompt.\n"
        else:
            msg += "Disabling new line before prompt.\n"
    msg += "Proceed?"

    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        with suspend_screen():
            apply_shell_config(changes)
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))

    # One last notification
    say_reload_terminal("Your shell config files were updated!")


# -------- MAIN SCREENS -------- #


def do_shell_screen():
    title = "Shell & Prompt"
    items = [
        ("Configure Prompt", "Configure the command prompt"),
        ("Default Login Shell", "Set the default login shell"),
        ("Reset Shell Config", "Reset the shell config files to their default"),
    ]
    buttons = [
        ("Select", "ok"),
        ("Back", "cancel"),
    ]
    ok_funcs = [
        do_shell_change_prompt,
        do_shell_change_default_login,
        do_shell_reset,
    ]

    while True:
        btn, line = SCREEN.show_listbox_window(title, items, buttons)
        if btn != "ok":
            break
        func = ok_funcs[line]
        while True:
            ret = func()
            if ret == "go-to-reset":
                func = do_shell_reset
                ret = func()
            if ret != "retry":
                break


def do_virtual_screen():
    title = "Virtualization"
    buttons = [
        ("Select", "ok"),
        ("Back", "cancel"),
    ]

    # Detect what kind of virtualization is that
    virt = VirtualizationSetting()
    virt.detect()
    logger.debug("Virtualization: %s", virt.type)

    # Not virtualized?
    if virt.type is None:
        msg = "Couldn't detect virtualization type."
        SCREEN.show_info_window("info", msg)
        return

    # Virtualized, but we don't provide any additional config?
    if not virt.is_configurable():
        msg = (
            f"Virtualization detected: {virt.pretty_name}.\n\n"
            "Nothing to configure for this virtualization technology."
        )
        SCREEN.show_info_window("info", msg)
        return

    # We might want to configure, or to unconfigure, or to do nothing,
    # in case the system is configured and there's no way to unconfigure.
    if virt.needs_configuration():
        items = [("Configure", virt.config_desc)]
        action = virt.configure
    elif virt.is_unconfigurable():
        items = [("Unconfigure", virt.undo_desc)]
        action = virt.unconfigure
    else:
        msg = (
            f"Virtualization detected: {virt.pretty_name}.\n\n"
            "Your system is already configured for this virtualization technology."
        )
        SCREEN.show_info_window("info", msg)
        return

    # Display the one-liner menu
    btn, _ = SCREEN.show_listbox_window(title, items, buttons)
    if btn != "ok":
        return

    # Do the actual job
    try:
        msg = (
            "For more information about Kali inside {virt.pretty_name}, please refer to:\n"
            f"{virt.help_url}"
        )
        with suspend_screen(ending_message=msg, wait_enter=True):
            action()
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))


def do_network_repo_screen():
    title = "Network Repository"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get the current apt config
    config = get_current_apt_sources_config()

    # Show the window with all the settings
    content = KaliNetworkRepoContent(config)
    btn = SCREEN.show_window(title, content, buttons)
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    msg = ""
    if "bleeding-edge" in changes:
        if changes["bleeding-edge"] is True:
            msg += "Enabling bleeding-edge repository.\n"
        else:
            msg += "Disabling bleeding-edge repository.\n"
    if "experimental" in changes:
        if changes["experimental"] is True:
            msg += "Enabling experimental repository.\n"
        else:
            msg += "Disabling experimental repository.\n"
    msg += "Proceed?"

    # Ask for confirmation
    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        msg = (
            "For more information about Kali repositories, please refer to:\n"
            "https://www.kali.org/docs/general-use/kali-linux-sources-list-repositories/"
        )
        with suspend_screen(ending_message=msg, wait_enter=True):
            apply_current_apt_sources_config(changes)
            apt_update()
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))


def do_hardening_screen():
    title = "Hardening Settings"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # All hardening settings
    hardening_settings = []

    # OpenSSL setting
    if is_pkg_installed("openssl"):
        setting = {
            "name": "openssl",
            "desc": "OpenSSL Configuration",
            "type": "radio",
            "choices": [
                (
                    "Wide Compatibility",
                    "Enable insecure protocols and ciphers, allowing access to legacy services",
                    "kali_wide_compatibility",
                ),
                (
                    "Strong Security",
                    "Only allow modern protocols and ciphers, reducing accessibility but increasing security",
                    "kali_strong_security",
                ),
            ],
            "value": None,
        }

        # Get the setting value
        try:
            openssl_setting = OpenSSLSetting()
            value = openssl_setting.get_ssl_system_default()
        except (KeyError, IOError) as e:
            msg = (
                "Failed to read the OpenSSL configuration!\n\n"
                "Please make sure that your Kali Linux installation"
                " is up to date and try again.\n\n"
                f"{e}"
            )
            SCREEN.show_info_window("error", msg)
            return

        # Assign everything
        setting["value"] = value
        hardening_settings.append(setting)

    # Show the window with all the settings
    content = KaliGenericContent(hardening_settings)
    btn = SCREEN.show_window(title, content, buttons)
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    msg = ""
    if "openssl" in changes:
        new = changes["openssl"]
        msg += f"Setting OpenSSL configuration to '{new}'.\n"
    msg += "Proceed?"

    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return "retry"

    # Do the actual job
    msg = (
        "For more information about OpenSSL configuration, please refer to:\n"
        "https://www.kali.org/docs/general-use/openssl-configuration/"
    )
    try:
        with suspend_screen(ending_message=msg, wait_enter=True):
            openssl_setting = OpenSSLSetting()
            openssl_setting.set_ssl_system_default(changes["openssl"])
    except (KeyError, IOError, ValueError) as e:
        msg = (
            "Failed to update the OpenSSL configuration!\n\n"
            "Please make sure that your Kali Linux installation"
            " is up to date and try again.\n\n"
            f"{e}"
        )
        SCREEN.show_info_window("error", msg)


def do_metapackages_screen():
    title = "Metapackages"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Check for kali-tools-* & kali-linux-*
    items = get_metapackage_items()

    # No Kali tools #1?
    if not items:
        with suspend_screen():
            apt_update()
        items = get_metapackage_items()

    # No Kali tools #2?
    if not items:
        msg = "No Kali Tools found. Are you really running Kali Linux? Or do you have an apt issue?"
        SCREEN.show_info_window("info", msg)
        return

    # Show the window with all the tools
    content = KaliSimpleCheckboxTree(HEIGHT, items)
    btn = SCREEN.show_window(title, content, buttons)
    if btn != "ok":
        return

    # Sum up changes
    selection = content.getSelection()
    to_install = []
    to_remove = []
    for pkg, _, installed in items:
        package = "kali-tools-" + pkg
        package = package.replace("kali-tools-kali-linux-", "kali-linux-")
        if pkg in selection and not installed:
            to_install += [package]
        elif pkg not in selection and installed:
            to_remove += [package]

    # No change?
    if to_install == [] and to_remove == []:
        return

    # Ask for confirmation
    msg = ""
    if to_install:
        msg += "Packages to install: "
        msg += ", ".join(to_install)
        msg += "\n"
    if to_remove:
        msg += "Packages to remove: "
        msg += ", ".join(to_remove)
        msg += "\n"
    msg += "Proceed?"

    btn = SCREEN.show_confirm_window(title, msg)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        msg = (
            "For more information about Kali's metapackages, please refer to:\n"
            "https://www.kali.org/docs/general-use/metapackages/"
        )
        with suspend_screen(ending_message=msg, wait_enter=True):
            if to_remove:
                remove_pkgs(to_remove)
            if to_install:
                install_pkgs(to_install)
    except RuntimeError as e:
        SCREEN.show_info_window("error", str(e))


def do_main_screen():
    title = "Main Menu"
    items = [
        ("Hardening", "Configure the system for extra security"),
        ("Metapackages", "Install specific subsets of tools for particular needs"),
        ("Network Repositories", "Configure network repositories for APT sources"),
        ("Shell & Prompt", "Configure the shell and command prompt"),
        ("Virtualization", "Additional configurations for Virtual Machines"),
    ]
    buttons = [
        ("Select", "ok"),
        ("Quit", "cancel"),
    ]
    ok_funcs = [
        do_hardening_screen,
        do_metapackages_screen,
        do_network_repo_screen,
        do_shell_screen,
        do_virtual_screen,
    ]

    while True:
        btn, line = SCREEN.show_listbox_window(title, items, buttons)
        if btn != "ok":
            break
        func = ok_funcs[line]
        while True:
            ret = func()
            if ret != "retry":
                break


def main():
    global SCREEN
    SCREEN = KaliTweaksScreen()
    try:
        do_main_screen()
    finally:
        SCREEN.destroy()
        # Print logs to stdout, if any
        if os.path.isfile(LOG_FILE):
            with open(LOG_FILE, "r") as f:
                print(f.read(), end="")
            os.remove(LOG_FILE)


if __name__ == "__main__":
    main()
