# Copyright 2021 Offensive Security
# SPDX-license-identifier: GPL-3.0-only

import errno
import filecmp
import getpass
import logging
import os
import re
import shutil
import subprocess
import tempfile
from pathlib import Path

logger = logging.getLogger(__name__)


def bold(text):
    bold = "\033[1m"
    reset = "\033[0m"
    return bold + text + reset


def say(text):
    print(bold(">>> " + text))
    logger.debug(text)


def say_install_packages(pkgs):
    if isinstance(pkgs, str):
        pkgs = [pkgs]
    if len(pkgs) == 0:
        text = "No package to install"
    elif len(pkgs) == 1:
        text = "Installing package: " + pkgs[0]
    else:
        text = "Installing packages: " + " ".join(pkgs)
    say(text)


def say_remove_packages(pkgs):
    if isinstance(pkgs, str):
        pkgs = [pkgs]
    if len(pkgs) == 0:
        text = "No package to remove"
    elif len(pkgs) == 1:
        text = "Removing package: " + pkgs[0]
    else:
        text = "Removing packages: " + " ".join(pkgs)
    say(text)


def say_install_program(program):
    say(f"Installing program: {program}")


def say_enable_service(service):
    say(f"Enabling service: {service}")


def say_disable_service(service):
    say(f"Disabling service: {service}")


def say_change_login_shell(shell):
    say(f"Setting login shell to '{shell}'")


def backup_filename(filename):
    """Return a backup filename"""
    return filename + ".kali-tweaks-orig"


def tmp_filename(filename):
    """Return a temporary filename"""
    # XXX Use proper tmp function
    return filename + ".tmp"


def _get_kali_tweaks_file(directory, filename):
    """Return the path to a kali-tweaks file"""
    if "RUN_FROM_SOURCE_TREE" in os.environ:
        f = os.path.join(os.getcwd(), directory, filename)
        if os.path.isfile(f):
            return f
    else:
        # path is hardcoded in setup.cfg, so it's hardcoded here as well
        f = os.path.join("/usr/lib/kali_tweaks", directory, filename)
        if os.path.isfile(f):
            return f
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), f)


def get_datafile(filename):
    """Return the path to a data file"""
    return _get_kali_tweaks_file("data", filename)


def get_helper(filename):
    """Return the path to a helper"""
    return _get_kali_tweaks_file("helpers", filename)


def _run(cmd, interactive=False):
    """If interactive is set, don't capture the output"""
    logger.debug("Running cmd: %s", cmd)
    res = subprocess.run(cmd, capture_output=(not interactive), shell=True, text=True)
    logger.debug("Return code: %d", res.returncode)
    # logger.debug("Stdout     :\n%s", res.stdout)
    # logger.debug("Stderr     :\n%s", res.stderr)
    return res


def run(cmd, **kwargs):
    assert not cmd.startswith("sudo")
    return _run(cmd, **kwargs)


def run_as_root(cmd, log=True, **kwargs):
    # For commands prefixed with sudo: interestingly there's no need to set
    # interactive=True, as sudo writes the prompt straight on the terminal
    # device, see sudo(8), option '-S'.
    if os.geteuid() == 0:
        rootcmd = cmd
    else:
        rootcmd = "sudo " + cmd
    if log is True:
        print("> $ " + rootcmd)
    return _run(rootcmd, **kwargs)


def apt_update():
    say("Updating APT cache")
    cmd = "apt-get update"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError("Couldn't update the APT cache")


def install_pkgs(pkgs, reinstall=False):
    apt_update()
    say_install_packages(pkgs)
    cmd = "apt-get -y install"
    if reinstall:
        cmd += " --reinstall"
    if isinstance(pkgs, list):
        pkgs = " ".join(pkgs)
    cmd += f" {pkgs}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install package(s) '{pkgs}'")


def remove_pkgs(pkgs):
    say_remove_packages(pkgs)
    cmd = "apt-get -y --purge remove"
    if isinstance(pkgs, list):
        pkgs = " ".join(pkgs)
    cmd += f" {pkgs}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't remove package(s) '{pkgs}'")


def is_pkg_installed(pkg):
    res = run(f"dpkg -s {pkg} 2>/dev/null | grep -q 'ok installed'")
    return res.returncode == 0


def install_program(program):
    say_install_program(program)
    dst = program
    src = get_datafile(os.path.basename(program))
    cmd = f"install -v -D -m0755 {src} {dst}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install program {src} -> {dst}")


def is_program_installed(program):
    if not os.path.isfile(program):
        return False
    if not os.access(program, os.X_OK):
        return False
    src = get_datafile(os.path.basename(program))
    return filecmp.cmp(src, program)


def enable_service_now(service):
    say_enable_service(service)
    cmd = f"systemctl start {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't start service '{service}'")
    cmd = f"systemctl enable {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't enable service '{service}'")


def disable_service_now(service):
    say_disable_service(service)
    cmd = f"systemctl stop {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't stop service '{service}'")
    cmd = f"systemctl disable {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't disable service '{service}'")


def is_service_up(service):
    cmd = f"systemctl is-active {service}"
    res = run(cmd)
    if res.stdout.strip() != "active":
        return False
    cmd = f"systemctl is-enabled {service}"
    res = run(cmd)
    if res.stdout.strip() != "enabled":
        return False
    return True


# -------- SHELL HELPERS -------- #


def shell_from_shell_path(shell_path):
    """Get a shell name from a shell path, ie: /bin/zsh -> zsh"""
    for sh in ["zsh", "bash", "fish", "ksh", "sh", "tcsh"]:
        if f"bin/{sh}" in shell_path:
            return sh
    return None


def shell_path_from_shell(shell):
    """Get a shell path from a shell name, ie: zsh -> /usr/bin/zsh"""
    if shell == "bash":
        return "/usr/bin/bash"
    elif shell == "zsh":
        return "/usr/bin/zsh"
    else:
        return None


def get_configured_login_shell():
    """Get the login shell, as configured in /etc/passwd"""
    username = getpass.getuser()
    res = run(f"getent passwd {username}")
    if res.returncode != 0:
        logger.debug("getent passwd %s: %s", username, res.returncode)
        return None

    line = res.stdout.strip()
    _, _, shell_path = line.rpartition(":")
    # per passwd(5), the last field might be empty,
    # in such case it defaults to bin/sh
    if not shell_path:
        shell_path = "/bin/sh"
    logger.debug("Detected shell_path: %s (from /etc/passwd)", shell_path)

    shell = shell_from_shell_path(shell_path)
    logger.debug("Corresponding shell: %s", shell)
    return shell


def get_effective_login_shell():
    """Get the current login shell, as given by the 'SHELL' variable"""
    shell_path = os.getenv("SHELL")
    if shell_path is None:
        logger.debug("shell_path is None")
        return None
    logger.debug("Detected shell_path: %s (from SHELL)", shell_path)

    shell = shell_from_shell_path(shell_path)
    logger.debug("Corresponding shell: %s", shell)
    return shell


def change_login_shell(sh):
    """Change the login shell (requires logout+login to be effective)"""
    say_change_login_shell(sh)
    shell_path = shell_path_from_shell(sh)
    if shell_path is None:
        raise NotImplementedError(f"Can't set user shell to: {sh}")
    cmd = f"chsh -s {shell_path}"
    res = run(cmd, interactive=True)
    if res.returncode != 0:
        msg = f"Couldn't set login shell to '{sh}' ({shell_path})\n"
        msg += "Check console output for details."
        raise RuntimeError(msg)


def get_shell_config_files():
    """Return a list of tuples (default, current)"""
    configs = []
    homedir = str(Path.home())
    skeldir = "/etc/skel"
    for rc in [".bashrc", ".zshrc"]:
        default = os.path.join(skeldir, rc)
        current = os.path.join(homedir, rc)
        if not os.path.isfile(default):
            continue
        if not os.access(default, os.R_OK):
            continue
        configs += [(default, current)]
    return configs


def apply_shell_config_files(configs):
    """Take a list of tuples from get_shell_config_files()"""
    say("Resetting shell config files")
    try:
        for default, current in configs:
            if os.path.isfile(current):
                backup = backup_filename(current)
                print(f"> '{current}' -> '{backup}'")
                shutil.copyfile(current, backup)
            print(f"> '{default}' -> '{current}'")
            shutil.copyfile(default, current)
    except OSError as e:
        raise RuntimeError(str(e))


def _get_rc_file(filename):
    homedir = str(Path.home())
    path = os.path.join(homedir, filename)
    if not os.path.isfile(path):
        logger.debug("shell_rc is not a file")
        return None
    if not os.access(path, os.R_OK):
        logger.debug("no access to shell_rc")
        return None
    return path


def _get_shell_rc_file(shell):
    if shell == "bash":
        return _get_rc_file(".bashrc")
    elif shell == "zsh":
        return _get_rc_file(".zshrc")
    logger.debug("unknown rc name")
    return None


def _get_kali_config_block(data):
    """The whole config block as lines, including delimiters"""
    start_delimiter = "# START KALI CONFIG VARIABLES"
    stop_delimiter = "# STOP KALI CONFIG VARIABLES"
    res = []
    keep = False
    for line in data.splitlines():
        if line == start_delimiter:
            res = [line]
            keep = True
        elif line == stop_delimiter:
            res += [line]
            keep = False
            break
        elif keep:
            res += [line]

    if keep is True:
        # there was a start delimiter, but no stop delimiter
        return []

    return res


def _get_shell_config_from_file(filename):
    """Retrieve some config values from the shell config file"""
    # Get the configuration block
    with open(filename) as f:
        data = f.read()
    content = _get_kali_config_block(data)
    if not content:
        logger.debug("config is empty")
        return None

    # Drop empty lines and comments
    content = [line.strip(" \t") for line in content]
    content = [line for line in content if line and not line.startswith("#")]

    # Make a dict out of that, drop anything unexpected.
    config = {}
    for line in content:
        key, _, value = line.partition("=")
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        elif value.startswith("'") and value.endswith("'"):
            value = value[1:-1]
        if key == "NEWLINE_BEFORE_PROMPT":
            if value in ["yes", "no"]:
                config["newline"] = value == "yes"
        elif key == "PROMPT_ALTERNATIVE":
            if value in ["oneline", "twoline", "backtrack"]:
                config["prompt"] = value

    return config


def _apply_shell_config_to_file(filename, config):
    """Apply some config values to the shell config file"""
    # Get the configuration block
    with open(filename) as f:
        data = f.read()
    content = _get_kali_config_block(data)
    if not content:
        return

    # Keep a copy for later
    content_orig = content.copy()

    # Set some config (add it if missing)
    for key, value in config.items():
        if key == "newline":
            shkey = "NEWLINE_BEFORE_PROMPT"
            value = "yes" if value else "no"
        elif key == "prompt":
            shkey = "PROMPT_ALTERNATIVE"
        else:
            continue

        print(f"> Setting: {shkey}='{value}'")

        changed = False
        for idx, line in enumerate(content):
            line = line.strip(" \t")
            if not line.startswith(shkey):
                continue
            content[idx] = f"{shkey}='{value}'"
            changed = True
            break

        if changed:
            continue

        content.append(f"{shkey}='{value}'")

    # Build a text block from content
    config_block_orig = "\n".join(content_orig)
    config_block = "\n".join(content)

    logger.debug("config orig:\n%s", config_block_orig)
    logger.debug("config     :\n%s", config_block)

    # Return if no changes
    if config_block_orig == config_block:
        return

    # Do the actual replacement
    data = data.replace(config_block_orig, config_block)

    # Write to file
    tmp = tmp_filename(filename)
    with open(tmp, "w") as f:
        f.write(data)

    backup = backup_filename(filename)
    shutil.copyfile(filename, backup)
    shutil.move(tmp, filename)


def get_current_shell_config():
    """
    Retrieve the config for the current shell. Beware that config can
    be an *empty dict* (config file found, but no config variables in
    there), which is different from *None* (no config file found).
    """
    shell = get_effective_login_shell()
    if not shell:
        logger.debug("No shell detected")
        return None
    logger.debug("Detected shell: %s", shell)

    rc_file = _get_shell_rc_file(shell)
    if not rc_file:
        logger.debug("No shell_rc detected")
        return None
    logger.debug("Detected shell_rc: %s", rc_file)

    return _get_shell_config_from_file(rc_file)


def apply_shell_config(config):
    """Apply the config to the supported shells"""
    say("Configuring command prompt")
    for shell in ["bash", "zsh"]:
        rc_file = _get_shell_rc_file(shell)
        if not rc_file:
            continue
        print(f"> Updating: {rc_file}")
        _apply_shell_config_to_file(rc_file, config)


# -------- APT REPO HELPERS -------- #


def _get_listfile(suite):
    return f"/etc/apt/sources.list.d/kali-{suite}.list"


def _remove_listfile(suite):
    listfile = _get_listfile(suite)
    print(f"> Removed repository: {listfile}")
    cmd = f"rm -fv {listfile}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't remove file '{listfile}'")


def _install_listfile(suite):
    listfile = _get_listfile(suite)
    mirror = "http://http.kali.org/kali"
    components = "main contrib non-free"
    content = f"deb {mirror} kali-{suite} {components}"
    print(f"> New repository: {content}")
    print(f"> Installing to: '{listfile}'")
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(content)
    cmd = f"install -m 0644 {f.name} {listfile}"
    res = run_as_root(cmd, interactive=True, log=False)
    os.remove(f.name)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install file to '{listfile}'")


def get_current_apt_sources_config():
    """Retrieve APT sources configuration"""
    config = {}
    for suite in ["bleeding-edge", "experimental"]:
        listfile = _get_listfile(suite)
        if os.path.isfile(listfile):
            config[suite] = True
        else:
            config[suite] = False
    return config


def apply_current_apt_sources_config(config):
    """Apply changes to the to APT sources"""
    say("Configuring APT repositories")
    for suite in ["bleeding-edge", "experimental"]:
        if suite not in config:
            continue
        if config[suite] is True:
            _install_listfile(suite)
        else:
            _remove_listfile(suite)


# -------- VIRTUALIZATION -------- #


class VirtualizationSetting:
    def __init__(self):
        self.type = None
        self.pretty_name = None
        self.config_desc = None
        self.undo_desc = None
        self.help_url = "https://www.kali.org/docs/virtualization/"
        self.packages = []  # packages to install
        self.scripts = []  # scripts to install
        self.config_script = None  # a config script to execute
        self.services = []  # systemd services to start

    def detect(self):
        res = run("systemd-detect-virt")
        if res.returncode != 0:
            return

        virt = res.stdout.strip()
        logger.debug("Detected virt: %s", virt)

        self.type = virt

        # There is no need to explicitly start the virt services, as they
        # are not "network services" (see [1]), therefore they are either:
        # 1. already started if they were already installed when the
        #    system was booted.
        # 2. started by apt when kali-tweaks installs the corresponding
        #    package.
        #
        # That's why there's no need to list it here. We only mention
        # services that are not auto-started and need an explicit start
        # and enable, like xrdp, which is indeed a network service.
        #
        # [1]: https://www.kali.org/docs/policy/kali-linux-network-service-policy/

        if virt in ["kvm", "qemu"]:
            self.pretty_name = virt.upper()
            self.config_desc = "Install additional packages for the SPICE protocol"
            self.packages = ["spice-vdagent", "xserver-xorg-video-qxl"]
        elif virt == "oracle":
            self.pretty_name = "VirtualBox"
            self.config_desc = "Install additional packages for VirtualBox"
            self.help_url += "install-virtualbox-guest-additions/"
            self.packages = ["virtualbox-guest-x11"]
        elif virt == "vmware":
            self.pretty_name = "VMWare"
            self.config_desc = "Install additional packages and scripts for VMware"
            self.help_url += "install-vmware-guest-tools/"
            self.packages = ["open-vm-tools-desktop"]
            self.scripts = [
                "/usr/local/sbin/mount-shared-folders",
                "/usr/local/sbin/restart-vm-tools",
            ]
        elif virt == "microsoft":
            self.pretty_name = "Hyper-V"
            self.config_desc = "Configure the system for Hyper-V enhanced session mode"
            self.undo_desc = "Remove configuration for Hyper-V enhanced session mode"
            self.help_url += "install-hyper-v-guest-enhanced-session-mode/"
            self.packages = ["hyperv-daemons", "xrdp"]
            self.config_script = get_helper("hyperv-enhanced-mode")
            self.services = ["xrdp"]
        else:
            logger.debug("Unsupported virt: %s", virt)
            self.type = None

    def is_configurable(self):
        """
        Return whether this virtualization tech is configurable.
        """
        if self.packages or self.scripts or self.config_script or self.services:
            return True
        return False

    def is_unconfigurable(self):
        """
        Return whether this virtualization tech can be unconfigured.
        """
        if self.config_script or self.services:
            return True
        return False

    def needs_configuration(self):
        """
        Return whether the system needs to be configured for this
        virtualization tech.
        """
        # Do we need to install packages?
        for pkg in self.packages:
            if not is_pkg_installed(pkg):
                return True

        # Do we need to install scripts?
        for script in self.scripts:
            if not is_program_installed(script):
                return True

        # Do we need to run a config script?
        if self.config_script:
            res = run(f"{self.config_script} check")
            if res.returncode != 0:  # unconfigured
                return True

        # Do we need to start a service?
        for svc in self.services:
            if not is_service_up(svc):
                return True

        return False

    def configure(self):
        """
        Configure for the virtualization type.

        This method raises RuntimeError if something goes wrong.
        """
        # Install missing packages
        packages = []
        for pkg in self.packages:
            if not is_pkg_installed(pkg):
                packages.append(pkg)
        if packages:
            install_pkgs(packages)

        # Install scripts
        for script in self.scripts:
            install_program(script)

        # Run config script
        if self.config_script:
            say(f"Configuring {self.pretty_name}")
            res = run_as_root(f"{self.config_script} enable", interactive=True)
            if res.returncode != 0:
                raise RuntimeError(
                    f"Couldn't configure the system for {self.pretty_name}"
                )

        # Enable services
        for svc in self.services:
            enable_service_now(svc)

    def unconfigure(self):
        """
        Unconfigure for the virtualization type.

        This is not perfect. For example, we stop services. But did
        we really started it, or was it started already before?
        """
        # Disable services
        for svc in reversed(self.services):
            disable_service_now(svc)

        # Run config script
        if self.config_script:
            say(f"Unconfiguring {self.pretty_name}")
            res = run_as_root(f"{self.config_script} disable", interactive=True)
            if res.returncode != 0:
                raise RuntimeError(
                    f"Couldn't unconfigure the system for {self.pretty_name}"
                )

        # Scripts and packages that might have been installed by the
        # configure() method are left behind.


# -------- OPENSSL -------- #


class OpenSSLConfigParser:
    """
    A simple parser for the OpenSSL configuration file.

    The goal is NOT to implement a full-fledged parser, but rather to
    provide only what's needed for our use-case:
    * read a config file
    * get a value within a section
    * change a value within a section
    * write back the file with our modifications

    When it comes to modifying the config file, we want to make sure to
    touch ONLY the line(s) that need modification, and leave the rest of
    the files untouched. That's why Python's configparser is not really
    suitable for our use-case (I'm not even sure it could parse the file
    anyway).

    Known limitations of this implementation:
    * in case a section/value would be duplicated, we only consider
      the first one. I think it's acceptable (I'm not even sure that
      duplicate sections are allowed).
    """

    def __init__(self, path):
        self.path = path
        self.config = None

    def read(self):
        """Read a config file, raise IOError on failure."""
        with open(self.path) as f:
            self.config = f.read().splitlines()

    def write(self):
        """Write a config file, raise IOError on failure."""
        print(f"> Writing changes to {self.path}")
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            for line in self.config:
                f.write(f"{line}\n")
        cmd = f"install -m 0644 {f.name} {self.path}"
        res = run_as_root(cmd, interactive=True, log=False)
        os.remove(f.name)
        if res.returncode != 0:
            raise RuntimeError(f"Couldn't install file to '{self.path}'")

    def _get_section(self, section):
        """
        Look for a section, and return the indexes of the first
        and the last line of this section.

        If the section is not found, KeyError is raised.
        """
        start = -1
        stop = -1
        within = False

        for idx, line in enumerate(self.config):
            # Not within the section?
            if not within:
                if re.match(fr"\[ *{section} *\]", line):
                    within = True
                    start = stop = idx
                continue

            # Within the section! End of the section?
            if line.startswith("["):
                break

            # This line belongs to the section
            stop = idx

        if start >= 0 and stop >= 0:
            return start, stop

        raise KeyError(f"In file {self.path}: no section {section}.")

    def _get_value(self, start, stop, key):
        """
        Look for a key between the lines start and stop. If found,
        return the index of the line, and the value.

        If the key is not found, KeyError is raised.

        Implementation details:
        * Looking at /etc/ssl/openssl.cnf, we can see that a trailing
          comment after a value is allowed.  And we can even see that
          there's no need for a space before the '#', it can be stuck
          to the value (there's actually such a line).
        * We assume that leading spaces before a key are not allowed.
        """
        for idx in range(start, stop + 1):
            line = self.config[idx]
            grp = re.match(f"({key}) *= *([^#]*)", line)
            if not grp:
                continue
            key, val = grp.groups()
            return idx, val.strip()

        raise KeyError(
            f"In file {self.path}: no key {key} between lines {start} and {stop}."
        )

    def _set_value(self, start, stop, key, value):
        """
        Set 'key' to 'value'. The key is looked for between the
        lines 'start' and 'stop', and must exist already.

        If the key does not exist, KeyError is raised.
        """
        index, _ = self._get_value(start, stop, key)
        self.config[index] = f"{key} = {value}"

    def get_value(self, section_name, key):
        """
        Get 'key' in 'section_name', raise KeyError if not found.
        """
        start, stop = self._get_section(section_name)
        _, value = self._get_value(start, stop, key)
        logger.debug("get [%s] %s: %s", section_name, key, value)
        return value

    def set_value(self, section_name, key, value):
        """
        Set 'key' in 'section_name' to 'value', raise KeyError if
        the key is not found.
        """
        start, stop = self._get_section(section_name)
        self._set_value(start, stop, key, value)
        logger.debug("set [%s] %s: %s", section_name, key, value)

    def get_section(self, section_name):
        """
        Get 'section_name', raise KeyError if not found.
        """
        start, stop = self._get_section(section_name)
        return self.config[start : stop + 1]

    def includes(self, filepath):
        """
        Return True is the file 'filepath' is included, False otherwise.

        Implementation details:
        * Per config(5), an optional '=' is allowed between the include
          directive and the path. No idea if trailing comments are allowed
          though. Let's just assume that if they are, there's a space before
          the '#'.
        """
        for line in self.config:
            if re.match(fr"\.include *=* *{filepath}(\s|$)", line):
                return True
        return False


class OpenSSLSetting:
    def __init__(self):
        self.kali_cnf = "/etc/ssl/kali.cnf"
        self.openssl_cnf = "/etc/ssl/openssl.cnf"

    def get_ssl_system_default(self):
        """
        Get the value for the setting '[ssl_sect] system_default'.

        This method raises IOError if the OpenSSL configuration file
        can't be read, and KeyError if the setting is not found.
        """
        conf = OpenSSLConfigParser(self.openssl_cnf)
        conf.read()
        value = conf.get_value("ssl_sect", "system_default")
        return value

    def set_ssl_system_default(self, value):
        """
        Set the value for the setting '[ssl_sect] system_default'.

        This method raises IOError if the OpenSSL configuration file
        can't be read or written, ValueError if the value is not found,
        and KeyError is the setting is not found.
        """

        say("Configuring OpenSSL")
        print(f"> Setting ssl system default to: {value}")

        # Make sure the value corresponds to a section in kali.cnf
        conf = OpenSSLConfigParser(self.kali_cnf)
        conf.read()
        try:
            _ = conf.get_section(value)
        except KeyError as e:
            raise ValueError(
                f"Invalid value '{value}', doesn't match a section.\n{e}"
            ) from e

        # Make sure the file kali.cnf is included
        conf = OpenSSLConfigParser(self.openssl_cnf)
        conf.read()
        if not conf.includes(self.kali_cnf):
            raise KeyError(f"File {self.openssl_cnf} does not include {self.kali_cnf}.")

        # Write the value to openssl.cnf
        conf.set_value("ssl_sect", "system_default", value)
        conf.write()
